 /*
 *    NaiveBayes.java
 *    Copyright (C) 2004 Bee-Chung Chen
 */

package LFA2;

import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.UpdateableClassifier;
import weka.filters.supervised.attribute.Discretize;
import java.io.*;
import java.util.*;
import weka.core.*;
import weka.estimators.*;

/**
 * Class for a Naive Bayes classifier using estimator classes. Numeric 
 * estimator precision values are chosen based on analysis of the 
 * training data. For this reason, the classifier is not an 
 * UpdateableClassifier (which in typical usage are initialized with zero 
 * training instances) -- if you need the UpdateableClassifier functionality,
 * use the NaiveBayesUpdateable classifier. The NaiveBayesUpdateable
 * classifier will  use a default precision of 0.1 for numeric attributes
 * when buildClassifier is called with zero training instances.
 * <p>
 * For more information on Naive Bayes classifiers, see<p>
 *
 * George H. John and Pat Langley (1995). <i>Estimating
 * Continuous Distributions in Bayesian Classifiers</i>. Proceedings
 * of the Eleventh Conference on Uncertainty in Artificial
 * Intelligence. pp. 338-345. Morgan Kaufmann, San Mateo.<p>
 *
 * Valid options are:<p>
 *
 * -K <br>
 * Use kernel estimation for modelling numeric attributes rather than
 * a single normal distribution.<p>
 *
 * -D <br>
 * Use supervised discretization to process numeric attributes.<p>
 *
 */
public class BaseNaiveBayes extends weka.classifiers.bayes.NaiveBayes {

  // In weka.classifiers.bayes.NaiveBayes
  //
  //    The attribute estimators:
  //    Estimator [][] m_Distributions;
  //
  //    The class estimator:
  //    Estimator m_ClassDistribution;

  public Estimator getClassPrior(){
    return m_ClassDistribution;
  }

  public double getClassPrior(int classValue){
    return m_ClassDistribution.getProbability(classValue);
  }

  /**
   * @return estimator[i][j]: prob(Attribute i | Class = j)
   */
  public Estimator[][] getAttrConditionals(){
    return m_Distributions;
  }

  /**
   * @return prob[i][j]: prob(Attribute i | Class = j, the instance)
   */
  public double[][] getAttrConditionals(Instance instance)
    throws Exception{
	Instance inst = (Instance) instance.copy();
    if (m_UseDiscretization) {
      m_Disc.input(inst);
      inst = m_Disc.output();
    }
    if(!m_Instances.equalHeaders(inst.dataset())){
        throw new Exception("Header mismatch!");
    }
    double[][] prob = new double[inst.numAttributes()][m_NumClasses];
    Enumeration enumAtts = inst.enumerateAttributes();
    int attIndex = 0;
    while (enumAtts.hasMoreElements()) {
      Attribute attribute = (Attribute) enumAtts.nextElement();
      for (int j = 0; j < m_NumClasses; j++) {
        if (!inst.isMissing(attribute)) {
          prob[attIndex][j] = m_Distributions[attIndex][j].
          getProbability(inst.value(attribute));
          if (Double.isNaN(prob[attIndex][j])) {
            throw new Exception("NaN returned from estimator for attribute "
				+ attribute.name() + ":\n"
				+ m_Distributions[attIndex][j].toString());
          }
        }else{
          prob[attIndex][j] = 0.1; // just use 0.1 for missing value
                                    // It doesn't matter, since for all class,
                                    // the probabilities are the same (0.1)
        }
      }
      attIndex++;
    }
    return prob;
  }

  public Instances getDatasetFormat(){
    return m_Instances;
  }

  public Discretize getDiscretizer(){
    return m_Disc;
  }
  
  /**
   * Main method for testing this class.
   *
   * @param argv the options
   */
  public static void main(String [] argv) {

    try {
      System.out.println(Evaluation.evaluateModel(new NaiveBayes(), argv));
    } catch (Exception e) {
      e.printStackTrace();
      System.err.println(e.getMessage());
    }
  }
}
