/*
 *    Ensemble.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */

package LFA2;

import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.core.Option;
import weka.core.Instances;
import weka.core.Instance;
import java.util.Vector;
import java.util.Enumeration;
import weka.classifiers.Evaluation;
import java.io.*;

/**
 * Base class for handling ensembles.
 *
 * This class, in fact, implements the Direct Ensemble method
 * of Learning From Aggregates.<p>
 *
 * Note: Class label is always the last attribute.
 *
 * Options:<p>
 *
 * -A aggregation schema file <br>
 *
 * -Z working directory <br>
 *
 * -B classifierstring or -B classifierstring1;classifierstring2;...<br>
 *    Classifierstring, e.g. "weka.classifier.trees.J48 -U", 
 *    should contain the full class name of a scheme
 *    included for selection followed by options to the classifier. <br>
 *    "-B classifierstring" means we build the same kind of classifiers
 *    on all the training tables.<br>
 *    "-B classifierstring1;classifierstring2;..." means we build classifierstring1
 *    on the 1st training table, and so on. The number of classifierstrings should
 *    be the same as the number of training tables.<p>
 *
 * -s sample-number (distinct number of samples per group) <br>
 *     1: just use average.
 *     0: use average without count
 *    -1: do not use average
 *
 * -r n
 *    Replicate each training example (after sampling) n times.
 *
 * -w weighting-method <br>
 *    Specify the weighting method (default: uniform).<p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */
public class Ensemble extends LFAClassifier {
  
  /** Array for storing the generated base classifiers. */
  protected Classifier[] m_Classifiers = null;

  /** weighting method */
  protected String m_WeightingMethod = null;
  
  /** Aggregation schema */
  protected SchemaInfo m_Schema = null;

  protected String m_WorkingDir = null;

  protected int m_SampleNum = 10;

  protected int m_ReplicateNum = 1;

  /**
   * Returns an enumeration describing the available options
   *
   * -A aggregation_schema_file <br>
   *
   * -Z working directory <br>
   *
   * -B classifierstring or -B classifierstring1;classifierstring2;...<br>
   * Classifierstring, e.g. "weka.classifier.trees.J48 -U", 
   * should contain the full class name of a scheme
   * included for selection followed by options to the classifier. <br>
   * "-B classifierstring" means we build the same kind of classifiers
   * on all the training tables.<br>
   * "-B classifierstring1;classifierstring2;..." means we build classifierstring1
   * on the 1st training table, and so on. The number of classifierstrings should
   * be the same as the number of training tables.
   * <p>
   *
   * -s sample-number (distinct number of samples per group) <br>
   *
   * -w weighting-method <br>
   * Specify the weighting method (default: uniform).<p>
   *
   * @return an enumeration of all the available options
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(6);

    newVector.addElement(new Option(
	      "\tFull class name of classifier to include, followed\n"
	      + "\tby scheme options. May be specified multiple times.\n"
	      + "\t(default: \"weka.classifiers.trees.J48\")",
	      "B", 1, "-B <classifier specification>"));
    newVector.addElement(new Option(
          "\t.E.g. -A test.schema",
          "A", 1, "-A <aggregation scheme file>"));
    newVector.addElement(new Option(
          "\t.E.g. -Z my/test",
          "Z", 1, "-Z <working directory>"));
    newVector.addElement(new Option(
          "\t.E.g. -s 10 (default:10)",
          "s", 1, "-s <sample number>"));
    newVector.addElement(new Option(
          "\t.E.g. -r 2 (default:1)",
          "r", 1, "-r <number of replicates>"));
    newVector.addElement(new Option(
          "\t.E.g. -w uniform (default: uniform)",
          "w", 1, "-w <weighting method>"));

    Enumeration enume = super.listOptions();
    while (enume.hasMoreElements()) {
      newVector.addElement(enume.nextElement());
    }
    return newVector.elements();
  }

  /**
   * Parses a given list of options. Valid options are:<p>
   *
   * -A aggregation_schema_file <p>
   *
   * -Z working directory <p>
   *
   * -B classifierstring or -B classifierstring1;classifierstring2;...<p>
   *
   * -s sample number <p>
   *
   * -r replicate number <p>
   *
   * -w weighting-method <p>
   *
   * @param options the list of options as an array of strings
   * @exception Exception if an option is not supported
   */
  public void setOptions(String[] options) throws Exception {

    //System.out.println("in setOptions");
    String check = Utils.getOption('c', options);
    if(check.length() > 0){
        throw new IllegalArgumentException(
                    "Class label is always the last attribute.\n"
                  ); 
    }
    
    String SchemaFileName = Utils.getOption('A', options);
    m_WorkingDir = Utils.getOption('Z', options);

    String temp = Utils.getOption('s', options);
    if(temp.length() != 0) {
        m_SampleNum = Integer.parseInt(temp);
    }

    temp = Utils.getOption('r', options);
    if(temp.length() != 0) {
        m_ReplicateNum = Integer.parseInt(temp);
    }
    
    temp = Utils.getOption('w', options);
    if(temp.length() != 0) {
        m_WeightingMethod = temp;
    }else{
        m_WeightingMethod = "uniform";
    }
    
    m_Schema = new SchemaInfo(SchemaFileName,m_WorkingDir,m_SampleNum,m_ReplicateNum);

    int numTables = m_Schema.numOfTables;
    m_Classifiers = new Classifier[numTables];

    // No option setting should after this part
    temp = Utils.getOption('B', options);
    if(temp.length() == 0){
        for(int i=0; i<numTables; i++){
            m_Classifiers[i] = new weka.classifiers.trees.J48();
        }
        return;
    }
    String[] strings = temp.split(";");
    if(strings.length == 1){
        String[] classifierSpec = Utils.splitOptions(strings[0]);
        String classifierName = classifierSpec[0];
        classifierSpec[0] = "";
        for(int i=0; i<numTables; i++){
            // MUST create a new array of options, because Utils.getOption(...)
            // called in forName(...) will eat options after it processed it.
            String[] opt = (String[]) classifierSpec.clone();
            m_Classifiers[i] = Classifier.forName(classifierName, opt);
        }
        return;        
    }
    if(strings.length != numTables){
        throw new IllegalArgumentException("The number of training tables and "+
                  "the number of base classifiers should be the same.\n"+
                  "Aggregation schema: "+SchemaFileName+"\n"+
                  "Classifier string: "+temp+"\n");
    }
    for(int i=0; i<numTables; i++){
        String[] classifierSpec = Utils.splitOptions(strings[i]);
        String classifierName = classifierSpec[0];
        classifierSpec[0] = "";
        m_Classifiers[i] = Classifier.forName(classifierName, classifierSpec);
    }
    // No code should be after this line.
    //System.out.println("end setOptions");
  }

  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    // System.out.println("In getOption()");
    String [] superOptions = super.getOptions();
    int current = 0;
    String[] options = new String [superOptions.length + 10];
    if(m_Schema != null){
        options[current++] = "-A";
        options[current++] = m_Schema.schemaFile;
    }
    if(m_WorkingDir != null){
        options[current++] = "-Z";
        options[current++] = m_WorkingDir;
    }

    options[current++] = "-s";
    options[current++] = m_SampleNum+"";

    options[current++] = "-r";
    options[current++] = m_ReplicateNum+"";

    if(m_WeightingMethod != null){
        options[current++] = "-w";
        options[current++] = m_WeightingMethod;
    }
    if(m_Classifiers != null){
        StringBuffer temp = new StringBuffer();
        for(int i=0; i<m_Classifiers.length; i++){
            if(i>=1) temp.append(";");
            temp.append(getClassifierSpec(i));
        }
        options[current++] = "-B";
        options[current++] = temp.toString();
    }

    System.arraycopy(superOptions, 0, options, current, 
                     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    // System.out.println("End getOption(): return ");
    // for(int i=0; i<options.length; i++){
    //     System.out.println("   "+options[i]);
    // }
    return options;
  }
  
  /**
   * Returns the tip text for this property
   * @return tip text for this property suitable for
   * displaying in the explorer/experimenter gui
   */
  public String classifiersTipText() {
    return "The base classifiers to be used.";
  }

  /**
   * Sets the list of possible classifers to choose from.
   *
   * @param classifiers an array of classifiers with all options set.
   */
  public void setClassifiers(Classifier [] classifiers) {

    m_Classifiers = classifiers;
  }

  /**
   * Gets the list of possible classifers to choose from.
   *
   * @return the array of Classifiers
   */
  public Classifier [] getClassifiers() {

    return m_Classifiers;
  }
  
  /**
   * Gets a single classifier from the set of available classifiers.
   *
   * @param index the index of the classifier wanted
   * @return the Classifier
   */
  public Classifier getClassifier(int index) {
    return m_Classifiers[index];
  }
  
  
  /**
   * Gets the classifier specification string, which contains the class name of
   * the classifier and any options to the classifier
   *
   * @param index the index of the classifier string to retrieve, starting from
   * 0.
   * @return the classifier string, or the empty string if no classifier
   * has been assigned (or the index given is out of range).
   */
  protected String getClassifierSpec(int index) {
    
    if (m_Classifiers.length < index) {
      return "";
    }
    Classifier c = getClassifier(index);
    return c.getClass().getName() + " "
      + Utils.joinOptions(((OptionHandler)c).getOptions());
  }
  
   /**
   * Buildclassifier builds all the base classifiers
   *
   * @param data the training data.
   * @exception Exception if the classifier could not be built successfully
   */
  public void buildClassifier() throws Exception {

    for(int i=0; i<m_Classifiers.length; i++){
      Instances data = m_Schema.getInstances(i);
      data.deleteWithMissingClass();
      //check_data(data);
      m_Classifiers[i].buildClassifier(data);
      data = null;
    }
  }
  
  void check_data(Instances data){
      for(int i=0; i<data.numInstances(); i++){
          Instance inst = data.instance(i);
          System.out.println(inst+" ("+inst.weight()+")");
      }
  }

  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    double [] sums = new double [instance.numClasses()], newProbs; 
    
    Instance[] testdata = m_Schema.separateInstance(instance);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }
    
    for (int i = 0; i < m_Classifiers.length; i++) {
      if (instance.classAttribute().isNumeric() == true) {
        sums[0] += m_Classifiers[i].classifyInstance(testdata[i]);
      } else {
        newProbs = m_Classifiers[i].distributionForInstance(testdata[i]);
        for (int j = 0; j < newProbs.length; j++)
          sums[j] += newProbs[j];
      }
    }
    if (instance.classAttribute().isNumeric() == true) {
      sums[0] /= (double)m_Classifiers.length;
      return sums;
    } else if (Utils.eq(Utils.sum(sums), 0)) {
      return sums;
    } else {
      Utils.normalize(sums);
      return sums;
    }
  }


  /**
   * Output a representation of this classifier
   */
  public String toString() {

    if (m_Classifiers == null) {
      return "LFA.Ensemble: No model built yet.";
    }

    String result = "=== Base Classifiers ===\n\n";
    for (int i = 0; i < m_Classifiers.length; i++) {
      result += "Base Classifier "+i+": ";
      result += m_Classifiers[i].toString() + "\n";
    }

    return result;
  }

  public String getName(){
      return "Direct Ensemble";
  }

  /** evaluate()
   *  output: name, its_accuracy, training_time, testing_time, base_accuracies ...
   */
  public void evaluate(String[] options) throws Exception {
    String test_file = weka.core.Utils.getOption('T', options);
    if("".equals(test_file)) throw new Exception("Please specify -T option");
    boolean printModel = weka.core.Utils.getFlag('P', options);
    BufferedReader in = new BufferedReader(new FileReader(test_file));
    Instances test_set = new Instances(in);
    test_set.setClassIndex(test_set.numAttributes()-1);
    in.close();
    setOptions(options);

    long begin_time = System.currentTimeMillis();
    buildClassifier();
    long end_time = System.currentTimeMillis();
    double train_time = ((double)(end_time - begin_time))/1000.0;

    if(printModel) System.out.println(this.toString());

    int correct = 0;
    int[] baseCorrect = new int[m_Schema.numOfTables];
    int i;
    for(i=0; i<m_Schema.numOfTables; i++) baseCorrect[i] = 0;

    begin_time = System.currentTimeMillis();
    for(i=0; i<test_set.numInstances(); i++){
      Instance tuple = test_set.instance(i);
      double label = classifyInstance(tuple);
      if(label == tuple.classValue()){
        correct++;
      }
    }
    end_time = System.currentTimeMillis();
    double test_time = ((double)(end_time - begin_time))/1000.0;

    for(i=0; i<test_set.numInstances(); i++){
      Instance tuple = test_set.instance(i);
      Instance[] tuples = m_Schema.separateInstance(tuple);
      for(int j=0; j<m_Schema.numOfTables; j++){
          double label = m_Classifiers[j].classifyInstance(tuples[j]);
          if(label == tuple.classValue()){
              baseCorrect[j]++;
          }
      }
    }

    double acc = ((double)correct)/((double)test_set.numInstances());
    System.out.print(getName()+", "+acc+", "+train_time+", "+test_time);
    for(i=0; i<m_Schema.numOfTables; i++){
        acc = ((double)baseCorrect[i])/((double)test_set.numInstances());
        System.out.print(", "+acc);
    }
    System.out.println("");
  }

  
  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * -t training file [-T test file] [-c class index]
   */
  public static void main(String [] argv) {
    Ensemble classifier = new Ensemble();
    try {
        classifier.evaluate(argv);
    } catch (Exception e) {
        e.printStackTrace();
        classifier.printOptions();
    }
  }

}
