/*
 *    NaiveBayes.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 */
 
 /**
 * The Naive Bayes for learning from aggregate views
 *
 * Options: (see also LFA.Ensemble) <p>
 *
 * -D debug-level<br>
 *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
 * <p>
 *
 * @author Bee-Chung Chen (beechung@cs.wisc.edu)
 */
 
// weka.classifiers.bayes.NaiveBayes will use initialize all counts by 1
 
package LFA2;
 
import LFA2.Ensemble;
import java.util.Iterator;
import weka.classifiers.Classifier;
import weka.core.OptionHandler;
import weka.core.Option;
import weka.core.Instances;
import weka.core.Instance;
import java.util.Vector;
import java.util.Enumeration;
import java.io.*;

public class NaiveBayes extends Ensemble {
    
  protected int debug = 0;
  
  double MIN_VALUE = 1E-20;
  
  protected BaseNaiveBayes[] m_baseNaiveBayes = null;

  /**
   * Returns an enumeration describing the available options
   * <p>
   * Options: (see also LFA.Ensemble) <p>
   *
   * -D debug-level<br>
   *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
   * <p>
   */
  public Enumeration listOptions() {

    Vector newVector = new Vector(1);

    newVector.addElement(new Option(
	      "\tdebug level: 1 to 5",
	      "D", 1, "-D <debug level>"));

    Enumeration enume = super.listOptions();
    while (enume.hasMoreElements()) {
      newVector.addElement(enume.nextElement());
    }
    return newVector.elements();
  }
  
  /**
   * Parses a given list of options. Valid options are:
   * <p>
   * Options: (see also LFA.Ensemble) <p>
   * 
   * -D debug-level<br>
   *    1: warning, 2: possible error, 3: interesting situation, 5: detailed message
   */
  public void setOptions(String[] options) throws Exception {

    String temp = weka.core.Utils.getOption('D', options);
    if(temp.length() == 0) debug = 0;
    else debug = Integer.parseInt(temp);

    super.setOptions(options);

    m_baseNaiveBayes = new BaseNaiveBayes[m_Classifiers.length];
    for(int i=0; i<m_Classifiers.length; i++){
        m_baseNaiveBayes[i] = (BaseNaiveBayes) m_Classifiers[i];
    }
  }
  
  /**
   * Buildclassifier builds all the base classifiers
   *
   * @param data the training data.
   * @exception Exception if the classifier could not be built successfully
   */
  public void buildClassifier(Instances data) throws Exception {
    super.buildClassifier(data);
  }
  

  /**
   * Calculates the class membership probabilities for the given test
   * instance.
   *
   * @param instance the instance to be classified
   * @return preedicted class probability distribution
   * @exception Exception if distribution can't be computed successfully 
   */
  public double[] distributionForInstance(Instance instance) throws Exception {

    if(instance.classAttribute().isNumeric() == true) {
    	throw new Exception("Cannot handle regression problem!!");
    }

    if(m_Schema.numOfAttrs-1 != m_Schema.classIndex)
        throw new Exception("(classIndex="+m_Schema.classIndex+") != (#Attrs="+m_Schema.numOfAttrs+") - 1");

    int nX = m_Schema.numOfAttrs - 1;
    int nY = instance.numClasses();
    int nViews = m_Schema.numOfTables;

    double[] prob = new double[nY];
    double[] pY = new double[nY];
    for(int i=0; i<pY.length; i++) pY[i] = 0;
    double[][] pXY = new double[nX][nY];
    for(int i=0; i<nX; i++) for(int j=0; j<nY; j++) pXY[i][j] = 0;

    Instance[] testdata = m_Schema.separateInstance(instance);
    if(testdata.length != m_Classifiers.length){
      throw new Exception("Number of classifiers ("+m_Classifiers.length+") "+
                          "=! number of training tables ("+testdata.length+")");
    }
    double[][][] VpXY = new double[nViews][][];
    for(int v=0; v<nViews; v++){
        VpXY[v] = m_baseNaiveBayes[v].getAttrConditionals(testdata[v]);
        for(int y=0; y<nY; y++) pY[y] += m_baseNaiveBayes[v].getClassPrior(y);
    }
    for(int y=0; y<nY; y++) pY[y] /= nViews;

    for(int y=0; y<nY; y++) for(int x=0; x<nX; x++){
        int count = 0;
        for(int v=0; v<nViews; v++){
            int colNo = m_Schema.baseColumnNo[x][v];
            if(colNo != -1){
                pXY[x][y] += VpXY[v][colNo][y];
                count++;
            }
        }
        if(count == 0) pXY[x][y]=1;
        else pXY[x][y] /= count;
        if(pXY[x][y] < MIN_VALUE) pXY[x][y] = MIN_VALUE;
    }

    for(int y=0; y<nY; y++){
        prob[y] = pY[y];
        for(int x=0; x<nX; x++) prob[y] *= pXY[x][y];
    }

    weka.core.Utils.normalize(prob);
    return prob;
  }


  /**
   * Output a representation of this classifier
   */
  public String toString() {
    if (m_Classifiers == null) {
      return "LFA2.NaiveBayes: No model built yet.";
    }
    StringBuffer result = new StringBuffer("=== Base Classifiers ===\n\n");
    for(int i=0; i<m_Classifiers.length; i++) {
      result.append("Base Classifier "+i+": ");
      result.append(m_Classifiers[i].toString() + "\n");
    }
    return result.toString();
  }

  
  /**
   * Gets the current settings of the Classifier.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String [] getOptions() {

    String [] superOptions = super.getOptions();
    int current = 0;
    String[] options = new String [superOptions.length + 2];
    if(debug > 0){
        options[current++] = "-D";
        options[current++] = "" + debug;
    }
    System.arraycopy(superOptions, 0, options, current, 
		     superOptions.length);
    current += superOptions.length;
    while (current < options.length) {
      options[current++] = "";
    }
    return options;
  }

  public String getName(){
      return "Naive Bayes";
  }

  /**
   * Main method for testing this class.
   *
   * @param argv should contain the following arguments:
   * 
   */
  public static void main(String [] argv) {
    NaiveBayes classifier = new NaiveBayes();
    try {
        classifier.evaluate(argv);
    } catch (Exception e) {
        e.printStackTrace();
        classifier.printOptions();
    }
  }
  
}
