/*
 *    SchemaInfo.java
 *    Copyright (C) 2004 Bee-Chung Chen
 *
 *    Schema file format:
 *    1st line: A file containing all the attributes (arff format)
 *    2nd line: the 1st training view
 *    3nd line: the 2nd training view
 *    ...
 *
 */

package LFA2;

import LFA2.*;
import java.io.*;
import java.util.LinkedList;
import weka.core.*;

public class SchemaInfo{

    public String schemaFile;

    TableSpec globalSpec;
    TableSpec[] baseSpec;

    String globalSpecFile;
    String[] baseSpecFile;

    int[] LFACountIndex; // -1 if no __LFACount__ column in the table

    int[][] globalAttrId;
    // globalAttrId[table_id][column_no] is the AttrId.
    // globalAttrId is column number of that attribute in the global table

    int[][] baseColumnNo;
    // baseColumnNo[globalAttrId][table_id]
    // -1 if the attribute does not in that table

    public int numOfAttrs;
    public int numOfTables;
    
    public int classIndex;

    public Instances getInstances(int table_id) throws Exception{
        if(table_id >= numOfTables)
            throw new Exception("table_id is out of bound\n");
        BufferedReader in = new BufferedReader(new FileReader(baseSpecFile[table_id]));
        Instances dataset = new Instances(in);
        dataset.setClassIndex(baseColumnNo[classIndex][table_id]);
        in.close();
        int countIndex = LFACountIndex[table_id];
        if(countIndex > 0){
            for(int i=0; i<dataset.numInstances(); i++){
                Instance tuple = dataset.instance(i);
                tuple.setWeight(tuple.value(countIndex));
                tuple.setValue(countIndex,0);
            }
        }
        return dataset;
    }

    public Instance[] separateInstance(Instance tuple){
        Instance[] tuples = new Instance[numOfTables];
        for(int i=0; i < numOfTables; i++){
            tuples[i] = new Instance(baseSpec[i].firstInstance());
            tuples[i].setDataset(baseSpec[i]);
            int countIndex = LFACountIndex[i];
            for(int j=0; j<baseSpec[i].numAttributes(); j++){
                if(countIndex == j){
                    tuples[i].setValue(j,0);
                }else{
                    tuples[i].setValue(j,tuple.value(globalAttrId[i][j]));
                }
            }
        }
        return tuples;
    }

    // the m-th tuple is of schema ordering[m]
    public Instance[] separateInstance(
        Instance tuple, int[] ordering, int[][] missingAttrIds
    ) throws Exception{
        Instance[] tuples = new Instance[ordering.length];
        if(missingAttrIds.length != ordering.length){
            throw new Exception("missing attribute list and aggregation scheme miss match ("+missingAttrIds.length+" vs "+ordering.length+")\n");
        }
        for(int m=0; m < ordering.length; m++){
            int i = ordering[m];
            tuples[m] = new Instance(baseSpec[i].firstInstance());
            tuples[m].setDataset(baseSpec[i]);
            int countIndex = LFACountIndex[i];
            int j;
            for(j=0; j<baseSpec[i].numAttributes(); j++){
                if(countIndex == j){
                    tuples[m].setValue(j,0);
                }else{
                    tuples[m].setValue(j,tuple.value(globalAttrId[i][j]));
                }
            }
            for(j=0; j<missingAttrIds[m].length; j++){
                // System.out.println("missing attr id:"+missingAttrIds[i][j]+" in table "+i);
                if(missingAttrIds[m][j] == -1) continue;
                tuples[m].setMissing(baseColumnNo[missingAttrIds[m][j]][i]);
            }
        }
        return tuples;
    }

    public SchemaInfo(String schema_filename, String workingDir, int sampleNum, int replicateNum) 
    throws Exception{
        String dir = "";
        schemaFile = schema_filename;
        if(workingDir != null && (!"".equals(workingDir))) dir = workingDir+"/";
        File schema_file = new File(schema_filename);
        if(!schema_file.exists())
            throw new Exception("Schema file '"+schema_filename+"' does not exist!\n");
        BufferedReader in = new BufferedReader(new FileReader(schema_file));
        String line = in.readLine().trim();
        if(line == null)
            throw new Exception("Schema file format error at line 1\n");
        File file = new File(dir+line);
        if(!file.exists())
            throw new Exception("File '"+workingDir+"/"+line+"' does not exist!\n");
        globalSpecFile = dir+line;
        LinkedList list = new LinkedList();
        while((line = in.readLine()) != null){
            line = line.trim();
            if(line.equals("")) continue;
            String[] tokens = line.split("\\s+");
            file = new File(dir+tokens[0]);
            if(!file.exists())
                throw new Exception("File '"+tokens[0]+"' does not exist!\n");
            list.add(dir+tokens[0]);
        }
        in.close();
        int nTables = list.size();
        if(nTables <= 0)
            throw new Exception("Schema file '"+schema_filename+"' contains no training table\n");
        Object[] temp = list.toArray();
        int i;
        baseSpecFile = new String[nTables];
        for(i=0; i<nTables; i++){
            baseSpecFile[i] = (String) temp[i];
        }
        in = new BufferedReader(new FileReader(globalSpecFile));
        globalSpec = new TableSpec(in);
        in.close();

        baseSpec = new TableSpec[nTables];
        for(i=0; i<baseSpec.length; i++){
            String originalFile = baseSpecFile[i];
            in = new BufferedReader(new FileReader(baseSpecFile[i]));
            baseSpec[i] = new TableSpec(in);
            in.close();
            if(ConvertToCount.needCovert(baseSpec[i])){
                ConvertToCount.convert(
                    baseSpecFile[i],baseSpecFile[i]+".count_view.arff",sampleNum
                );
                baseSpecFile[i] += ".count_view.arff";
            }
            if(replicateNum <= 0)
                throw new Exception("Cannot replicate training example "+replicateNum+" times\n");
            else if(replicateNum > 1){
                ConvertToCount.replicate(
                    baseSpecFile[i],baseSpecFile[i]+".replicated.arff",replicateNum
                );
                baseSpecFile[i] += ".replicated.arff";
            }
            if(!baseSpecFile[i].equals(originalFile)){
                in = new BufferedReader(new FileReader(baseSpecFile[i]));
                baseSpec[i] = new TableSpec(in);
                in.close();
            }
        }
        int nAttr = globalSpec.numAttributes();
        globalAttrId = new int[nTables][];
        baseColumnNo = new int[nAttr][nTables];
        for(i=0; i<nAttr; i++){
            for(int j=0; j<nTables; j++) baseColumnNo[i][j] = -1;
        }
        LFACountIndex = new int[nTables];
        for(i=0; i<nTables; i++) LFACountIndex[i]=-1;
        for(i=0; i<nTables; i++){
            if(baseSpec[i].numAttributes()>nAttr+3)
                throw new Exception("The "+i+"-th base table has too many attributes\n");
            globalAttrId[i] = new int[baseSpec[i].numAttributes()];
            for(int j=0; j<baseSpec[i].numAttributes(); j++){
                Attribute baseAttr = baseSpec[i].attribute(j);
                if("__LFACount__".equals(baseAttr.name())){
                    LFACountIndex[i] = j;
                    globalAttrId[i][j] = -1;
                    continue;
                }
                Attribute globalAttr = globalSpec.attribute(baseAttr.name());
                if(globalAttr == null)
                    throw new Exception("Attribute "+baseAttr.name()+" does not appear in global spec\n");
                if(!globalAttr.equals(baseAttr))
                    throw new Exception("Attribute "+baseAttr.name()+" is different from the global spec\n");
                globalAttrId[i][j] = globalAttr.index();
                baseColumnNo[globalAttr.index()][i] = j;
            }
        }
        numOfAttrs = nAttr;
        numOfTables = nTables;
        classIndex = numOfAttrs-1;
        for(i=0; i<nTables; i++)
            baseSpec[i].setClassIndex(baseColumnNo[classIndex][i]);
    }

    static void check(String schema_filename, String workingDir, int sampleNum) throws Exception{
        SchemaInfo schema = new SchemaInfo(schema_filename, workingDir, sampleNum, 1);
        System.out.println("Number of Tables: "+schema.numOfTables);
        System.out.println("Number of Attributes: "+schema.numOfAttrs);
        SchemaInfo temp = new SchemaInfo(schema_filename, workingDir, sampleNum, 1);
        for(int i=0; i<schema.numOfTables; i++){
            Instances data = schema.getInstances(i);
            BufferedReader in = new BufferedReader(new FileReader(temp.baseSpecFile[i]));
            Instances dataset = new Instances(in);
            dataset.setClassIndex(temp.baseColumnNo[temp.classIndex][i]);
            in.close();
            int countIndex = temp.LFACountIndex[i];
            if(countIndex > 0){
                for(int j=0; j<dataset.numInstances(); j++){
                    Instance tuple1 = dataset.instance(j);
                    Instance tuple2 = data.instance(j);
                    double[] t1 = tuple1.toDoubleArray();
                    double[] t2 = tuple2.toDoubleArray();
                    for(int k=0; k<t1.length; k++){
                        if(countIndex == k) continue;
                        if(t1[k] != t2[k]) System.out.println("ERROR");
                    }
                    if(t1[countIndex] != tuple2.weight())
                        System.out.println("ERROR");
                    //System.out.println(t1[countIndex]+" vs "+tuple2.weight());
                }
            }
        }
    }

    public static void main(String [] argv) throws Exception {
        check(argv[0], argv[1], Integer.parseInt(argv[2]));
    }
}
